unit SillyNetworkingU;

{$implicitExceptions off}

interface

uses
  types, sysutils, blcksock, synsock, Classes, syncobjs;

const
  MaxMessageSize = 1024 * 1024;
  MessageBufferSize = 64;

type
  TCustomSharedObject = class {$ifdef production} refcounted {$endif}
  end;

  TSharedObject = class(TCustomSharedObject)
  public
    constructor Create; virtual;
    function DebugOutputOnCreateEnabled: Boolean; virtual;
    function DebugOutputOnDestroyEnabled: Boolean; virtual;
    procedure WriteDebug(s: string);
    destructor Destroy; override;
  end;

  TCustomCriticalSection = class {$ifdef production} refcounted {$endif} (TCriticalSection)
  end;

  TCriticalSection = class(TCustomCriticalSection)
  end;

  TCustomMemory = class {$ifdef production} refcounted {$endif} (TMemoryStream)
  end;

  TMemory = class(TCustomMemory)
  protected
    function GetAsString: string;
    procedure SetAsString(a: string);
  public
    constructor Create;
    function CheckIfEmpty: Boolean;
    property AsString: string read GetAsString write SetAsString;
    property Empty: Boolean read CheckIfEmpty;
    destructor Destroy; override;
  end;

  TMemoryDynArray = array of TMemory;

  TCustomSocket = class {$ifdef production} refcounted {$endif} (TTCPBlockSocket)
  end;

  TSocket = class(TCustomSocket)
  public
    constructor Create; reintroduce;
    function DebugOutputOnCreateEnabled: Boolean; virtual;
    function DebugOutputOnDestroyEnabled: Boolean; virtual;
    destructor Destroy; override;
  end;

  TBuffer = class(TSharedObject)
  private
    MemoryF: TMemory;
    SizeF: Int64;
    ExpectedSize: Integer;
    procedure Shrink;
    function GetMemory: TMemory;
    procedure SetMemory(a: TMemory);
    // Create memory.
    // Does not assign Memory property.
    // Does not assign MemoryF field.
    // Returns new memory.
    function CreateMemory: TMemory;
    property Memory: TMemory read GetMemory write SetMemory;
    // Attempt to extract next data block.
    // Returns nil if block is not ready yet.
    function Extract: TMemory;
    function ExtractExpectedSize: Integer;
  public
    property Size: Int64 read SizeF write SizeF;
    constructor Create; override;
    // Returns -1 if expected size is not present.
    function WriteByte(a: Byte): TMemory;
    destructor Destroy; override;
  end;

  TMemoryQueue = class(TSharedObject)
  protected
    Data: TMemoryDynArray;
    Count: Integer;
    Locker: TCriticalSection;
    // Shitf items left and decrease count;
    procedure ShiftLeft;
  public
    constructor Create; override;
    function Push(a: TMemory): Boolean;
    function Pop: TMemory;
    function CheckIfEmpty: Boolean;
    destructor Destroy; override;
  end;

  TCustomMethodThread = class {$ifdef production} refcounted {$endif}(TThread)
  end;

  TMethodThread = class(TCustomMethodThread)
  protected
    procedure WriteDebug(s: string);
  public type TSillyMethod = procedure(aTerminated: PBoolean) of object;
  public
    Method: TSillyMethod;
    constructor Create(aMethod: TSillyMethod); reintroduce;
    procedure Execute; override;
    destructor Destroy; override;
  end;

  TKeepAlive = class(TSharedObject)
  protected
    LastMoment: QWord;
    Interval: QWord;
  public
    constructor Create; override;
    // returns True when it is time to send keep-alive data block.
    function CheckForward: Boolean;
    class function GetDefaultInterval: QWord;
  end;

  TClient = class(TSharedObject)
  public const
    LogKeepAlive = True;
  protected
    InboxBuffer: TBuffer;
    SocketF: TSocket;
    ReaderThread: TMethodThread;
    WriterThread: TMethodThread;
    KeepAlive: TKeepAlive;
    procedure RunReaderThread(aTerminated: PBoolean);
    procedure ReadForward;
    procedure ConnectForward;
    procedure RunWriterThread(aTerminated: PBoolean);
    procedure WriteForward;
    procedure SendKeepAliveIfRequired;
    procedure Start;
    procedure Stop;
    procedure Write(a: TMemory);
    function CheckSocketActive: Boolean;
  public
    ConnectionActive: Boolean;
    Inbox: TMemoryQueue;
    Outgoing: TMemoryQueue;
    Address: string;
    Port: Word;
    constructor Create; override;
    procedure Attach(aSocket: synsock.TSocket);
    procedure Connect;
    function Read(out aByte: Byte): Boolean;
    destructor Destroy; override;
  end;

  TClientDynArray = array of TClient;

  TServer = class(TSharedObject)
  protected
    ServerSocketF: TSocket;
    PortF: Word;
    Thread: TMethodThread;
    Clients: TClientDynArray;
    ClientsLocker: TCriticalSection;
    procedure RunThread(aTerminated: PBoolean);
    procedure CycleForward;
    procedure ListenAccept;
    procedure RemoveInactive;
    procedure Add(aClient: TClient);
  public
    property ServerSocket: TSocket read ServerSocketF;
    property Port: Word read PortF write PortF;
    constructor Create; override;
    function Start: Boolean;
    // Get list of currently present clients.
    // Use to iterate through clients.
    function GetClients: TClientDynArray;
    procedure Stop;
    destructor Destroy; override;
  end;

procedure WriteDebug(s: string);
function CopyExistingClients(a: TClientDynArray): TClientDynArray;

implementation

const
  SillyThreadStackSize = 64 * 1024;
  // Milliseconds.
  ThreadIdleInterval = 100;
  MaxClients = 128;

var
  GlobalSharedObjectCount: Integer;
  GlobalMemoryCount: Integer;

// Function not optimized.
function JoinStringArray(a: TStringDynArray; aSeparator: string): string;
var
  i, n: Integer;
begin
  n := Length(a);
  if n > 0 then
  begin
    result := a[0];
    for i := 1 to n - 1 do
      result := result + aSeparator + a[i];
  end
  else
    result := '';
end;

function ExceptionCallStackToStrings: TStringDynArray;
var
  i: Integer;
  frames: PPointer;
begin
  SetLength(result, 1 + ExceptFrameCount);
  frames := ExceptFrames;
  result[0] := BackTraceStrFunc(ExceptAddr);
  for i := 0 to ExceptFrameCount - 1 do
    result[i + 1] := BackTraceStrFunc(frames[i]);
end;

function ExceptionCallStackToText: string;
begin
  result := JoinStringArray(ExceptionCallStackToStrings, LineEnding);
end;

function ExceptionToText(e: Exception): string;
begin
  result := e.ClassName + ': "' + e.Message + '"' + LineEnding + ExceptionCallStackToText;
end;

// Write Debug Instance Create Destroy
procedure WriteDebugInstanceCD(aInstance: TObject; aCreate: Boolean);
var
  prefix: string;
begin
  if aCreate then
    prefix := 'Creating instance'
  else
    prefix := 'Destroying instance';
  WriteDebug(prefix + ': ' + aInstance.UnitName + '.' + aInstance.ClassName);
end;

{ TKeepAlive }

constructor TKeepAlive.Create;
begin
  inherited Create;
  LastMoment := GetTickCount64;
  Interval := GetDefaultInterval;
end;

function TKeepAlive.CheckForward: Boolean;
var
  currentTime: QWord;
  shouldSend: Boolean;
begin
  currentTime := GetTickCount64;;
  shouldSend := currentTime - LastMoment >= Interval;
  if shouldSend then
    LastMoment := currentTime;
  result := shouldSend;
end;

class function TKeepAlive.GetDefaultInterval: QWord;
begin
  result := 1000;
end;

{ TMethodThread }

procedure TMethodThread.WriteDebug(s: string);
begin
  SillyNetworkingU.WriteDebug(self.ClassName + ': ' + s);
end;

constructor TMethodThread.Create(aMethod: TSillyMethod);
begin
  inherited Create(True, SillyThreadStackSize);
  Method := aMethod;
  WriteDebugInstanceCD(self, True);
  Start;
end;

procedure TMethodThread.Execute;
begin
  try
    if Method <> nil then
      Method(@self.Terminated);
  except
    on e: Exception do
      WriteDebug('Exception while executing server thread: ' + ExceptionToText(e));
  end;
end;

destructor TMethodThread.Destroy;
begin
  WriteDebug('Now destroying, stopping');
  Terminate;
  WaitFor;
  WriteDebugInstanceCD(self, False);
  inherited Destroy;
end;

{ TMemoryQueue }

procedure TMemoryQueue.ShiftLeft;
var
  i: Integer;
begin
  for i := 0 to Count - 2 do
    Data[i] := Data[i + 1];
  Data[Count - 1] := nil;
  Dec(Count);
end;

constructor TMemoryQueue.Create;
begin
  inherited Create;
  SetLength(Data, MessageBufferSize);
  Count := 0;
  Locker := TCriticalSection.Create;
end;

function TMemoryQueue.Push(a: TMemory): Boolean;
begin
  Locker.Enter;
  result := Count < Length(Data);
  if result then
  begin
    Data[Count] := a;
    Inc(Count);
  end;
  Locker.Leave;
end;

function TMemoryQueue.Pop: TMemory;
begin
  Locker.Enter;
  if Count > 0 then
  begin
    result := Data[0];
    ShiftLeft;
  end
  else
    result := nil;
  Locker.Leave;
end;

function TMemoryQueue.CheckIfEmpty: Boolean;
begin
  Locker.Enter;
  result := 0 = Count;
  Locker.Leave;
end;

destructor TMemoryQueue.Destroy;
begin
  inherited Destroy;
end;

{ TMemory }

function TMemory.GetAsString: string;
begin
  SetLength(result, self.Size);
  if self.Size > 0 then
    Move(self.Memory^, result[1], Length(result));
end;

procedure TMemory.SetAsString(a: string);
begin
  self.Size := Length(a);
  self.Position := 0;
  if Length(a) > 0 then
    self.Write(a[1], Length(a));
end;

constructor TMemory.Create;
begin
  inherited Create;
  InterLockedIncrement(GlobalMemoryCount);
end;

function TMemory.CheckIfEmpty: Boolean;
begin
  result := Size = 0;
end;

destructor TMemory.Destroy;
begin
  InterLockedDecrement(GlobalMemoryCount);
  inherited Destroy;
end;

{ TBuffer }

function TBuffer.Extract: TMemory;
begin
  if ExpectedSize > -1 then
    if Memory.Position >= Int64(SizeOf(Integer)) + ExpectedSize then
    begin
      WriteDebug('Now extracting: ' + IntToStr(ExpectedSize));
      result := TMemory.Create;
      Memory.Position := SizeOf(Integer);
      if ExpectedSize > 0 then
        result.CopyFrom(Memory, ExpectedSize);
      Shrink;
      ExpectedSize := -1;
    end
    else
      result := nil;
end;

function TBuffer.ExtractExpectedSize: Integer;
var
  position: Int64;
begin
  result := -1;
  position := Memory.Position;
  Memory.Position := 0;
  Memory.ReadBuffer(result, SizeOf(result));
  Memory.Position := position;
end;

function TBuffer.CreateMemory: TMemory;
begin
  result := TMemory.Create;
  result.Size := Size;
  result.Position := 0;
end;

procedure TBuffer.Shrink;
begin
  MemoryF := CreateMemory;
end;

function TBuffer.GetMemory: TMemory;
begin
  if nil = MemoryF then
    MemoryF := CreateMemory;
  result := MemoryF;
end;

procedure TBuffer.SetMemory(a: TMemory);
begin
  MemoryF := a;
end;

constructor TBuffer.Create;
begin
  inherited Create;
  ExpectedSize := -1;
end;

function TBuffer.WriteByte(a: Byte): TMemory;
begin
  Memory.WriteByte(a);
  if ExpectedSize = -1 then
    if Memory.Position >= SizeOf(Integer) then
    begin
      ExpectedSize := ExtractExpectedSize;
    end;
  result := Extract;
end;

destructor TBuffer.Destroy;
begin
  MemoryF.Free;
  inherited Destroy;
end;

{ TClient }

procedure TClient.RunReaderThread(aTerminated: PBoolean);
begin
  while not aTerminated^ do
  begin
    ConnectForward;
    ReadForward;
    Sleep(ThreadIdleInterval);
  end;
  ReadForward;
end;

procedure TClient.ReadForward;
var
  lByte: Byte;
  block: TMemory;
begin
  while ConnectionActive and Read(lByte) do
  begin
    block := InboxBuffer.WriteByte(lByte);
    if block <> nil then
      if block.Size > 0 then
      begin
        WriteDebug('New message arrived');
        Inbox.Push(block)
      end
      else
        if LogKeepAlive then
          WriteDebug('Keep alive data block received');
    ConnectionActive := CheckSocketActive;
    if not ConnectionActive then
    begin
      WriteDebug('ReadForward: Disconnected while reading; '
        + 'error description = "' + SocketF.LastErrorDesc + '"');
      SocketF.CloseSocket;
    end;
  end;
end;

procedure TClient.ConnectForward;
begin
  if not ConnectionActive then
    if (Address <> '') and (Port <> 0) then
    begin
      WriteDebug('Now connecting to: ' + 'address: "' + Address + '", port: ' + IntToStr(Port));
      SocketF.Connect(Address, IntToStr(Port));
      if SocketF.LastError = 0 then
      begin
        WriteDebug('Connected.');
        ConnectionActive := True;
      end
      else
      begin
        WriteDebug('Error on connection: ' + SocketF.LastErrorDesc);
        SocketF.CloseSocket;
        ConnectionActive := False;
      end;
    end;
end;

procedure TClient.RunWriterThread(aTerminated: PBoolean);
begin
  while not aTerminated^ do
  begin
    WriteForward;
    Sleep(ThreadIdleInterval);
  end;
  WriteForward;
end;

procedure TClient.WriteForward;
var
  message: TMemory;
begin
  while ConnectionActive do
  begin
    //SendKeepAliveIfRequired;
    message := Outgoing.Pop;
    if message <> nil then
      Write(message);
    ConnectionActive := CheckSocketActive;
    if not ConnectionActive then
    begin
      WriteDebug('WriteForward: Disconnected while writing; '
        + 'error description: "' + SocketF.LastErrorDesc + '"');
      SocketF.CloseSocket;
    end;
    if message = nil then
      break;
  end;
end;

procedure TClient.SendKeepAliveIfRequired;
begin
  if KeepAlive.CheckForward then
  begin
    if LogKeepAlive then
      WriteDebug('Now sending keep alive data block');
    Write(TMemory.Create);
  end;
end;

procedure TClient.Start;
begin
  ReaderThread := TMethodThread.Create(@RunReaderThread);
  WriterThread := TMethodThread.Create(@RunWriterThread);
end;

procedure TClient.Stop;
begin
  ReaderThread := nil;
  WriterThread := nil;
end;

procedure TClient.Write(a: TMemory);
var
  lSize: Integer;
begin
  lSize := a.Size;;
  SocketF.SendBuffer(@lSize, SizeOf(Integer));
  SocketF.SendBuffer(a.Memory, lSize);
end;

function TClient.CheckSocketActive: Boolean;
begin
  result := (SocketF.LastError = 0) or (SocketF.LastError = WSAETIMEDOUT);
end;

constructor TClient.Create;
begin
  inherited Create;
  Inbox := TMemoryQueue.Create;
  InboxBuffer := TBuffer.Create;
  InboxBuffer.Size := MaxMessageSize + 1;
  Outgoing := TMemoryQueue.Create;
  KeepAlive := TKeepAlive.Create;
end;

procedure TClient.Attach(aSocket: synsock.TSocket);
begin
  SocketF := TSocket.Create;
  SocketF.Socket := aSocket;
  SocketF.GetSins;
  ConnectionActive := True;
  Start;
end;

procedure TClient.Connect;
begin
  SocketF := TSocket.Create;
  Start;
end;

function TClient.Read(out aByte: Byte): Boolean;
begin
  aByte := SocketF.RecvByte(1);
  result := SocketF.LastError = 0;
end;

destructor TClient.Destroy;
begin
  Stop;
  inherited Destroy;
end;


{ TSocket }

constructor TSocket.Create;
begin
  inherited Create;
  if DebugOutputOnCreateEnabled then
    WriteDebugInstanceCD(self, True);
end;

function TSocket.DebugOutputOnCreateEnabled: Boolean;
begin
  result := True;
end;

function TSocket.DebugOutputOnDestroyEnabled: Boolean;
begin
  result := True;
end;

destructor TSocket.Destroy;
begin
  if DebugOutputOnDestroyEnabled then
    WriteDebugInstanceCD(self, False);
  inherited Destroy;
end;

{ TServer }

procedure TServer.RunThread(aTerminated: PBoolean);
begin
  while not aTerminated^ do
  begin
    CycleForward;
    Sleep(ThreadIdleInterval);
  end;
  CycleForward;
end;

procedure TServer.CycleForward;
begin
  ListenAccept;
  RemoveInactive;
end;

procedure TServer.ListenAccept;
var
  client: TClient;
begin
  while ServerSocket.CanRead(ThreadIdleInterval) do
  begin
    client := TClient.Create;
    client.Attach(ServerSocket.Accept);
    Add(client);
  end;
end;

procedure TServer.RemoveInactive;
var
  i: Integer;
begin
  ClientsLocker.Enter;
  for i := 0 to Length(Clients) - 1 do
    if Clients[i] <> nil then
      if not Clients[i].ConnectionActive and Clients[i].Inbox.CheckIfEmpty then
        Clients[i] := nil;
  ClientsLocker.Leave;
end;

procedure TServer.Add(aClient: TClient);
var
  i: Integer;
begin
  ClientsLocker.Enter;
  for i := 0 to Length(Clients) - 1 do
    if nil = Clients[i] then
    begin
      Clients[i] := aClient;
      break;
    end;
  ClientsLocker.Leave;
end;

constructor TServer.Create;
begin
  inherited Create;
  SetLength(Clients, MaxClients);
  ClientsLocker := TCriticalSection.Create;
end;

function TServer.Start: Boolean;
begin
  ServerSocketF := TSocket.Create;
  ServerSocket.CreateSocket;
  ServerSocket.SetLinger(True, 10000);
  ServerSocket.Bind('0.0.0.0', IntToStr(Port));
  ServerSocket.Listen;
  result := ServerSocket.LastError = 0;
  WriteDebug('Server start result: ' + BoolToStr(result, True));
  if result then
    Thread := TMethodThread.Create(@RunThread);
end;

function TServer.GetClients: TClientDynArray;
begin
  ClientsLocker.Enter;
  result := CopyExistingClients(Clients);
  ClientsLocker.Leave;
end;

destructor TServer.Destroy;
begin
  Stop;
  inherited Destroy;
end;

procedure TServer.Stop;
begin
  Thread := nil;
end;

{ TSharedObject }

constructor TSharedObject.Create;
begin
  inherited Create;
  InterLockedIncrement(GlobalSharedObjectCount);
  if DebugOutputOnCreateEnabled then
    WriteDebugInstanceCD(self, True);
end;

function TSharedObject.DebugOutputOnCreateEnabled: Boolean;
begin
  result := True;
end;

function TSharedObject.DebugOutputOnDestroyEnabled: Boolean;
begin
  result := True;
end;

procedure TSharedObject.WriteDebug(s: string);
begin
  SillyNetworkingU.WriteDebug(self.ClassName + ': ' + s);
end;

destructor TSharedObject.Destroy;
begin
  InterLockedDecrement(GlobalSharedObjectCount);
  if DebugOutputOnDestroyEnabled then
    WriteDebugInstanceCD(self, False);
  inherited Destroy;
end;

{ TNoClass }

procedure WriteDebug(s: string);
begin
  WriteLN('SillyNetworking: ' + s);
end;

function CopyExistingClients(a: TClientDynArray): TClientDynArray;
var
  i, n: Integer;
begin
  n := 0;
  for i := 0 to Length(a) - 1 do
    if a[i] <> nil then
      Inc(n);
  SetLength(result, n);
  n := 0;
  for i := 0 to Length(a) - 1 do
    if a[i] <> nil then
    begin
      result[n] := a[i];
      Inc(n);
    end;
end;

finalization
  WriteDebug('Count of shared objects: ' + IntToStr(GlobalSharedObjectCount));
  WriteDebug('Count of memory: ' + IntToStr(GlobalSharedObjectCount));
end.

